#
# * ***** BEGIN GPL LICENSE BLOCK *****
# *
# * This program is free software; you can redistribute it and/or
# * modify it under the terms of the GNU General Public License
# * as published by the Free Software Foundation; either version 2
# * of the License, or (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software Foundation,
# * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

# * Contributor(s): Johannes Kollmer - www.polykraft.de
# *
# * ***** END GPL LICENSE BLOCK *****
# 

bl_info = {
    "name": "Blembic Importer",
    "author": "Johannes Kollmer (Lumpengnom)",
    "version": (1, 0),
    "blender": (2, 7, 6),
    "location": "Properties > Object Tab",
    "description": "Blembic is a Meshcache Plugin for efficiently streaming data from other 3D programs",
    "category": "Import-Export",
}

import bpy
import struct
import os
import mathutils
import gzip
import math

bpy.app.handlers.frame_change_post.clear()
bpy.app.handlers.frame_change_pre.clear()
bpy.app.handlers.scene_update_pre.clear()
bpy.app.handlers.render_pre.clear()
bpy.app.handlers.render_post.clear

# Assign a collection and add variables for user input such as the Y-Up checkbox
class SceneSettingItem(bpy.types.PropertyGroup):
    name = bpy.props.StringProperty(name="", default="Unknown",subtype ='DIR_PATH')
    BlembicMatList = bpy.props.StringProperty(name, default="Unknown")
    value = bpy.props.IntProperty(name="Test Prop Value", default=22)
    mystring = bpy.props.StringProperty(name="Test Prop mystring", default="Unknown", subtype ='DIR_PATH')
    mybool = bpy.props.BoolProperty(name="On/Off", default = True)
    mybool2 = bpy.props.BoolProperty(name="Normals", default = False)
    mybool3 = bpy.props.BoolProperty(name="Y-Up", default = False)
    mybool4 = bpy.props.BoolProperty(name="Mat", default = False)
# Create input fields like checkboxes in the UI    
class My_UL_TestList(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
        
        custom_icon = 'OBJECT_DATAMODE'
        mat_icon = 'MATERIAL'

        row = layout.row()
        box = layout.box()
        row2 = box.row()
        col = box.column()
        row3 = box.row()
        rowselect = layout.row()
        
        col.scale_x = 1
        col.scale_y = .7
        row.scale_x = 1.5
        row.scale_y = .7
        row2.scale_x = 1
        row2.scale_y = .7
        row3.scale_x = 1
        row3.scale_y = .7

        col.prop(item,"name",emboss=True,icon_value=0)              
        row.label(item.mystring, item.name, icon = custom_icon)
        row2.label(item.BlembicMatList, item.name, icon = mat_icon)
        
        
        row3.prop(item,"mybool")
        row3.prop(item,"mybool2")
        row3.prop(item, "mybool3")       
        row3.prop(item,"mybool4")
        #row.label(item.mystring)
        #row.label(item.name)
        #row.prop(context.scene,"OutputPath")


class MyPanelList(bpy.types.Panel):
    bl_label = "Blembic Mesh Cache"
    bl_idname = "blembic_importer_panel"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = 'object'
    
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        row = layout.row()
        row.template_list("My_UL_TestList", "", scene, "my_settings", scene, "my_settings_index")  
        row = layout.row()
        row.operator("addobject.button")
        row.operator("removeobject.button")
        row.operator("updateobject.button")
        row.operator("addmat.button")


class FolderButton(bpy.types.Operator):
    bl_idname = "folder.button"
    bl_label = "Select Folder"
    
    def execute(self, context):
        print ("MyTestButton")
        
        GroupExists = False
        
        for i in bpy.context.object.users_group:
            if i.name == "MyTestGroup2":
                GroupExists = True
                
        if GroupExists == True:
            #MyIndex = bpy.context.scene.my_settings_index
            #MyIndex = bpy.context.
            SelectedObj = bpy.context.object.name
            MyIndex = bpy.data.groups['MyTestGroup2'].objects.find(SelectedObj)            
            
            bpy.context.scene.my_settings[MyIndex].name = bpy.context.scene.OutputPath
                      
        return{'FINISHED'}    

class AddObjectButton(bpy.types.Operator):
    bl_idname = "addobject.button"
    bl_label = "Add Object"
    
    def execute(self, context):
        
        GroupExists = False
        
        BlembicList = bpy.context.scene.my_settings
        MyActiveObj = bpy.context.active_object
        
        for i in range (len(BlembicList)):
                        
            if MyActiveObj.name == BlembicList[i].mystring:
                GroupExists = True
            '''
            else:
                GroupExists = False
            '''
        if GroupExists == False:
            my_item = bpy.context.scene.my_settings.add()
            my_item.mystring = MyActiveObj.name
            my_item.name = "Select Folder"
        else:
            print ("already in group")
        return{'FINISHED'}    

class RemoveObjectButton(bpy.types.Operator):
    
    bl_idname = "removeobject.button"
    bl_label = "Remove Object"
    
    def execute(self,context):
        
        BlembicList = bpy.context.scene.my_settings
        MyActiveIndex = bpy.context.scene.my_settings_index

        bpy.context.scene.my_settings.remove(MyActiveIndex)
        return{'FINISHED'}


class UpdateObjectsButton(bpy.types.Operator):

    bl_idname = "updateobject.button"
    bl_label = "Update Object"
    
    def execute(self,context):

        GroupExists = False
        MyActiveObj = bpy.context.active_object
        BlembicList = bpy.context.scene.my_settings
        MyActiveIndex  = bpy.context.scene.my_settings_index

        for i in range (len(BlembicList)):
            if MyActiveObj.name == BlembicList[i].mystring:
                GroupExists = True
        
        if GroupExists == False:
            bpy.context.scene.my_settings[MyActiveIndex].mystring = MyActiveObj.name
        else:
            print("Can not update. Object allready in List")
        
        return{'FINISHED'}

class AddMatButton(bpy.types.Operator):
    bl_idname = "addmat.button"
    bl_label = "Add Materials"
    
    def execute(self, context):
        
        
        BlembicList = bpy.context.scene.my_settings
        MyActiveObj = bpy.context.active_object
                   
        my_item = bpy.context.scene.my_settings[bpy.context.scene.my_settings_index]
        my_item.BlembicMatList = MyActiveObj.name
        
        return{'FINISHED'}    


'''

def MyupdateObjects(dummy):

    for j in bpy.context.scene.my_settings:
        print (j.name)
'''

# The following is the script that runs every frame and loads the next Blembic file

def MyUpdateMesh(scene):

    # Get the List storing the Blembic Objects defined by the user    
    BlembicMeshList = bpy.context.scene.my_settings

    # Main Loop. Loops through all Blembic objects
    for BlembicObject in BlembicMeshList:
        
        # Check if object is on
        ObjectOnOffBool = (BlembicObject.mybool)
        
        # If Object is on update it, else do nothing.
        if ObjectOnOffBool == True:
            
            # Get Path where mesh is stored
            MyFilePath = bpy.path.abspath(BlembicObject.name)
            # Blembic Cache files have to be named fluidsurface_final_
            MyFileName = "fluidsurface_final_"
            MyFileNumber = str(bpy.context.scene.frame_current)
            MyFileNumberPad = "{0:0>4}".format(MyFileNumber)
            MyFileNamePost = ".BBC"
            print(MyFilePath+MyFileName+MyFileNumberPad+MyFileNamePost)
            MyFileSource = MyFilePath+MyFileName+MyFileNumberPad+MyFileNamePost
            
            FileExists = os.path.isfile(MyFileSource)
            
            # Check if File Exists
            if FileExists == True:
                MyVertList = []
                MyFaceList = []
                MySplitNormalsList = []
                MyMatIDList = []
                MyMapVertPosList = []
                MyMapFaceVertsList = []
                
                # Set filepath as variable and open the file. "rb" means "read binary"
                myfilename3 = bpy.path.abspath(MyFileSource)
                myfile3 = open(myfilename3,'rb')

                # Get number of vertices in file
                myread = myfile3.read(4)
                myNumVerts = (struct.unpack('i', myread))
                #print("Numverts")
                #print (myNumVerts[0])

                # Get Vertex Position
                for i in range(myNumVerts[0]):
                    myread2 = myfile3.read(12)
                    MyVertVector = []
                    floatstring = struct.unpack("fff", myread2)
                    MyVertList.append (floatstring)
                
                # reads number of vertices. Irrelevant. Should be removed in the future.
                myread3 = myfile3.read(4)

                # Get number of faces
                myread5 = myfile3.read(4)
                myNumPolys = (struct.unpack('i', myread5))

                # Check if user has tagged the Y-Up button
                print("YUp on or off?")
                YUpBool = (BlembicObject.mybool3)
                print(YUpBool)
                
                # Get vertex order for face creation. If Y-Up is checked get the third number first
                for i in range(myNumPolys[0]):
                    MyFaceVector=[]
                    myread6 = myfile3.read(12)
                    floatstring = struct.unpack("iii", myread6)
                    if YUpBool == False:
                        MyFaceList.append (floatstring)
                    else:
                        MyFaceVector.append(floatstring[0])
                        MyFaceVector.append(floatstring[2])
                        MyFaceVector.append(floatstring[1])
                        MyFaceList.append (MyFaceVector)
                        
                # get number of split normals
                myread7 = myfile3.read(4)
                myNumSplitNormals = (struct.unpack('i', myread7))
                print ("myNumSplitNormals")
                print (myNumSplitNormals[0])

                # check if user has checked split normals button
                SplitNormalsBool = (BlembicMeshList[0].mybool2)
                
                # Get Split Normals Vectors
                if myNumSplitNormals != 0:
                    for i in range(myNumSplitNormals[0]):
                        myread8 = myfile3.read(12)
                        if SplitNormalsBool == True:
                            floatstring = struct.unpack("fff", myread8)
                            tempfloatstring =[]
                            tempfloatstring.append (floatstring[0])
                            tempfloatstring.append (floatstring[1])
                            tempfloatstring.append (floatstring[2])
                            MySplitNormalsList.append (tempfloatstring)   

                # Get Material IDs per face
                for i in range(myNumPolys[0]):
                    myread8 = myfile3.read(4)
                    MyMatID =  struct.unpack('i', myread8)[0]
                    MyMatIDList.append(MyMatID)
                
                # Get number of Material IDs 
                myread9 = myfile3.read(4)
                MyNumMatIDs = struct.unpack("i", myread9)
                
                # Get number of UV-Maps
                myread10 = myfile3.read(4)
                MyNumMaps = struct.unpack("i", myread10)     
                
                # Get number of UV-Vertices
                myread11 = myfile3.read(4)
                MyNumMapVerts = struct.unpack('i', myread11)
                
                # Get UV-Vertex Positions
                for i in range(MyNumMapVerts[0]):
                    myread12 = myfile3.read(8)
                    MyMapVertPos = struct.unpack("ff", myread12)
                    MyMapVertPosList.append(MyMapVertPos) 
                #print ("MyMapVertPosList")
                #print (MyMapVertPosList)  
                
                # Get Number of UV-Faces                
                myread13 = myfile3.read(4)
                MyNumMapFaces = struct.unpack("i",myread13)

                # Cycle through UV-Faces and get their UV-Vertices' Positons  
                print ("MyNumMapFaces")
                print (MyNumMapFaces)              
                for i in range(MyNumMapFaces[0]):
                    myread14 = myfile3.read(12)
                    MyMapFaceVerts = struct.unpack("iii", myread14)
                    MyMapFaceVertsList.append(MyMapFaceVerts)

                # close the file    
                myfile3.close
                
                
                # Get the object in the scene via its name
                MyObject = bpy.data.objects[BlembicObject.mystring]
                
                # Get the mesh and create a new one with "from_pydata".
                # Split Normals require Autosmooth to be on
                MyOldMesh = MyObject.data
                MyName = MyObject.name
                MyMesh = bpy.data.meshes.new(MyName)
                MyMesh.from_pydata(MyVertList,[],MyFaceList)
                MyMesh.update()
                MyObject.data = MyMesh
                MyMesh.use_auto_smooth = True
                
                print ("My SplitNormals True or False")
                print (SplitNormalsBool)
                print (len(MyMesh.loops))

                #Check if user has ticked the Split Normals checkbox and add them if he has
                if SplitNormalsBool == True:
                    print("SplitNBool is true")
                    print (len(MyMesh.loops))
                    print (len(MySplitNormalsList))
                    if len(MyMesh.loops) == len(MySplitNormalsList):
                        print ("adding splitNormals")
                        MyMesh.normals_split_custom_set(MySplitNormalsList)
                
                # check user has assigned a material object and in case he has add the materials
                if bpy.data.objects.get(BlembicObject.BlembicMatList) is not None:
                    for i in range(len(bpy.data.objects[BlembicObject.BlembicMatList].data.materials)):
                        MyMesh.materials.append(bpy.data.objects[BlembicObject.BlembicMatList].data.materials[i])
                for i in range(len(MyMatIDList)):
                    MyMesh.polygons[i].material_index = MyMatIDList[i] - 1
                
                # Create UV-Maps
                print ("MyNumMaps")
                print (MyNumMaps)
                for i in range(MyNumMaps[0]):

                    MyMesh.loops.data.uv_textures.new()
                    MyCounter = 0
                    #print ("MyNumFaceVertsList")
                    #print (len(MyMapFaceVertsList))
                    for j in range(MyNumMapFaces[0]):
                        MyMapCorners = MyMapFaceVertsList[j]
                        MyMesh.uv_layers[i].data[MyCounter].uv = MyMapVertPosList[MyMapCorners[0]]
                        MyCounter = MyCounter + 1
                        MyMesh.uv_layers[i].data[MyCounter].uv = MyMapVertPosList[MyMapCorners[1]]
                        MyCounter = MyCounter + 1
                        MyMesh.uv_layers[i].data[MyCounter].uv = MyMapVertPosList[MyMapCorners[2]]
                        MyCounter = MyCounter + 1 
                    
                MyMesh.update()
                bpy.data.meshes.remove(MyOldMesh)


def register():
    bpy.utils.register_class(SceneSettingItem)
    bpy.utils.register_class(My_UL_TestList)
    bpy.utils.register_class(MyPanelList)
    bpy.utils.register_class(FolderButton)
    bpy.utils.register_class(AddObjectButton)    
    bpy.utils.register_class(AddMatButton)    
    bpy.utils.register_class(RemoveObjectButton)
    bpy.utils.register_class(UpdateObjectsButton)
    
    bpy.types.Scene.my_settings = bpy.props.CollectionProperty(type=SceneSettingItem)    
    bpy.types.Scene.my_settings_index = bpy.props.IntProperty(name = "Index for my List", default = 0)
    bpy.types.Scene.OutputPath = bpy.props.StringProperty(name="", subtype='DIR_PATH')
    #bpy.app.handlers.frame_change_pre.append(MyupdateObjects)
    bpy.app.handlers.persistent(MyUpdateMesh)
    bpy.app.handlers.frame_change_pre.append(MyUpdateMesh)
     
def unregister():
    bpy.app.handlers.frame_change_pre.remove(MyUpdateMesh)   
    #bpy.app.handlers.frame_change_pre.remove(MyupdateObjects)   
    del bpy.types.Scene.OutputPath
    del bpy.types.Scene.my_settings_index
    del bpy.types.Scene.my_settings
   
    bpy.utils.unregister_class(UpdateObjectsButton)  
    bpy.utils.unregister_class(RemoveObjectButton)  
    bpy.utils.unregister_class(AddMatButton)  
    bpy.utils.unregister_class(AddObjectButton)  
    bpy.utils.unregister_class(FolderButton)    
    bpy.utils.unregister_class(MyPanelList)
    bpy.utils.unregister_class(My_UL_TestList)    
    bpy.utils.unregister_class(SceneSettingItem)

    
    
if __name__ == "__main__":
    register()  